"""Objects for labelling and succesively running functions.
"""

import os
import shutil

import xarray as xr

from .prepare import (
    _parse_fn_args,
    _parse_var_names,
    _parse_var_dims,
    _parse_var_coords,
    _parse_constants,
    _parse_resources,
    _parse_combos,
    _parse_cases,
    _parse_attrs,
)
from .combo_runner import combo_runner_to_ds
from .case_runner import case_runner_to_ds
from .batch import Crop
from ..manage import load_ds, save_ds


# --------------------------------------------------------------------------- #
#                                   RUNNER                                    #
# --------------------------------------------------------------------------- #

class Runner(object):
    """Container class with all the information needed to systematically
    run a function over many parameters and capture the output in a dataset.

    Parameters
    ----------
    fn : callable
        Function that produces a single instance of a result.
    var_names : str, sequence of str, or None
        The ordered name(s) of the ouput variable(s) of `fn`. Set this
        explicitly to None if `fn` outputs already labelled data as a
        :class:`~xarray.Dataset` or :class:`~xarray.DataArray`.
    fn_args : str, or sequence of str, optional
        The ordered name(s) of the input arguments(s) of `fn`. This is only
        needed if the cases or combos supplied are not dict-like.
    var_dims : dict-like, optional
        Mapping of output variables to their named internal dimensions, can be
        the names of ``constants``.
    var_coords : dict-like, optional
        Mapping of output variables named internal dimensions to the actual
        values they take.
    constants : dict-like, optional
        Constants arguments to be supplied to `fn`. These can be used as
        'var_dims', and will be saved as coords if so, otherwise as attributes.
    resources : dict-like, optional
        Like `constants` but not saved to the the dataset, e.g. if very big.
    attrs : dict-like, optional
        Any other miscelleous information to be saved with the dataset.
    default_runner_settings
        These keyword arguments will be supplied as defaults to any runner.
    """

    def __init__(self, fn, var_names,
                 fn_args=None,
                 var_dims=None,
                 var_coords=None,
                 constants=None,
                 resources=None,
                 attrs=None,
                 **default_runner_settings):
        self.fn = fn
        self._var_names = _parse_var_names(var_names)
        self._fn_args = (_parse_fn_args(fn_args) if fn_args is not None else
                         None)
        self._var_dims = _parse_var_dims(var_dims, self._var_names)
        self._var_coords = _parse_var_coords(var_coords)
        self._constants = _parse_constants(constants)
        self._resources = _parse_resources(resources)
        self._attrs = _parse_attrs(attrs)
        self.default_runner_settings = default_runner_settings

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    # Attributes which should be parsed ------------------------------------- #

    def _get_fn_args(self):
        return self._fn_args

    def _set_fn_args(self, fn_args):
        self._fn_args = _parse_fn_args(fn_args)

    def _del_fn_args(self):
        self._fn_args = None
    fn_args = property(_get_fn_args, _set_fn_args, _del_fn_args,
                       "List of the names of the arguments that the "
                       "Runner's function takes.")

    def _get_var_names(self):
        return self._var_names

    def _set_var_names(self, var_names):
        self._var_names = _parse_var_names(var_names)

    def _del_var_names(self):
        self._var_names = None
    var_names = property(_get_var_names, _set_var_names, _del_var_names,
                         "List of the names of the variables that the "
                         "Runner's function produces.")

    def _get_var_dims(self):
        return self._var_dims

    def _set_var_dims(self, var_dims, var_names=None):
        if var_names is None:
            var_names = self._var_names
        self._var_dims = _parse_var_dims(var_dims, var_names)

    def _del_var_dims(self):
        self._var_dims = None
    var_dims = property(_get_var_dims, _set_var_dims, _del_var_dims,
                        "Mapping of each output variable to its named "
                        "dimensions")

    def _get_var_coords(self):
        return self._var_coords

    def _set_var_coords(self, var_coords):
        self._var_coords = _parse_var_coords(var_coords)

    def _del_var_coords(self):
        self._var_coords = None
    var_coords = property(_get_var_coords, _set_var_coords, _del_var_coords,
                          "Mapping of each variable named dimension to its "
                          "coordinate values.")

    def _get_constants(self):
        return self._constants

    def _set_constants(self, constants):
        self._constants = _parse_constants(constants)

    def _del_constants(self):
        self._constants = None
    constants = property(_get_constants, _set_constants, _del_constants,
                         "Mapping of constant arguments supplied to the "
                         "Runner's function.")

    def _get_resources(self):
        return self._resources

    def _set_resources(self, resources):
        self._resources = _parse_resources(resources)

    def _del_resources(self):
        self._resources = None
    resources = property(_get_resources, _set_resources, _del_resources,
                         "Mapping of constant arguments supplied to the "
                         "Runner's function that are *not* saved with the "
                         "dataset.")

    # Running methods ------------------------------------------------------- #

    def run_combos(self, combos, constants=(), **runner_settings):
        """Run combos using the function map and save to dataset.

        Parameters
        ----------
        combos : tuple of form ((str, seq), ...)
            The values of each function argument with which to evaluate all
            combinations.
        constants : dict (optional)
            Extra constant arguments for this run, repeated arguments will
            take precedence over stored constants but for this run only.
        runner_settings
            Keyword arguments supplied to :func:`~xyzpy.combo_runner`.
        """
        combos = _parse_combos(combos)
        self.last_ds = combo_runner_to_ds(
            self.fn, combos, self._var_names,
            var_dims=self._var_dims,
            var_coords=self._var_coords,
            constants={**self._constants, **dict(constants)},
            resources=self._resources,
            attrs=self._attrs,
            parse=False,
            **{**self.default_runner_settings, **runner_settings})
        return self.last_ds

    def run_cases(self, cases, constants=(), **runner_settings):
        """Run cases using the function and save to dataset.

        Parameters
        ----------
        cases : sequence of mappings or tuples
            A sequence of cases.
        constants : dict (optional)
            Extra constant arguments for this run, repeated arguments will
            take precedence over stored constants but for this run only.
        runner_settings
            Supplied to :func:`~xyzpy.case_runner`.
        """
        cases = _parse_cases(cases)
        self.last_ds = case_runner_to_ds(
            fn=self.fn,
            fn_args=self._fn_args,
            cases=cases,
            var_names=self._var_names,
            var_dims=self._var_dims,
            var_coords=self._var_coords,
            constants={**self._constants, **dict(constants)},
            resources=self._resources,
            attrs=self._attrs,
            parse=False,
            **{**self.default_runner_settings, **runner_settings})
        return self.last_ds

    def Crop(self,
             name=None,
             parent_dir=None,
             save_fn=None,
             batchsize=None,
             num_batches=None):
        """Return a Crop instance with this runner, from which ``fn``
        will be set, and then combos can be sown, grown, and reaped into the
        ``Runner.last_ds``. See :class:`~xyzpy.Crop`.

        Returns
        -------
        Crop
        """
        return Crop(runner=self,
                    name=name,
                    parent_dir=parent_dir,
                    save_fn=save_fn,
                    batchsize=batchsize,
                    num_batches=num_batches)

    def __repr__(self):
        string = "<xyzpy.Runner>\n"
        string += "    fn: {self.fn}\n"

        if self.fn_args is not None:
            string += "    fn_args: {self.fn_args}\n"

        string += "    var_names: {self.var_names}\n"

        if self.var_dims is not None:
            string += "    var_dims: {self.var_dims}\n"

        if self.constants:
            self._constants_list = list(self.constants)
            string += "    constants: {self._constants_list}\n"

        if self.resources:
            self._resources_list = list(self.resources)
            string += "    resources: {self._resources_list}\n"

        return string.format(self=self)


# --------------------------------------------------------------------------- #
#                                 HARVESTER                                   #
# --------------------------------------------------------------------------- #

class Harvester(object):
    """Container class for collecting and aggregating data to disk.

    Parameters
    ----------
    runner : Runner
        Performs the runs and describes the results.
    data_name : str, optional
        Base file path to save data to.
    chunks : int or dict, optional
        If not None, passed to xarray so that the full dataset is loaded and
        merged into with on-disk dask arrays.
    engine : str, optional
        Engine to use to save and load datasets.
    full_ds : xarray.Dataset, optional
        Initialize the Harvester with this dataset as the intitial full
        dataset.

    Members
    -------
    full_ds : xarray.Dataset
        Dataset containing all data harvested so far, by default synced to
        disk.
    last_ds : xarray.Dataset
        Dataset containing just the data from the last harvesting run.
    """

    def __init__(self, runner, data_name=None, chunks=None,
                 engine='h5netcdf', full_ds=None):
        self.runner = runner
        self.data_name = data_name
        self.engine = engine
        self.chunks = chunks
        self._full_ds = full_ds

    @property
    def last_ds(self):
        """Dataset containing the last runs' data.
        """
        return self.runner.last_ds

    def load_full_ds(self, chunks=None, engine=None):
        """Load the disk dataset into ``full_ds``.

        Parameters
        ----------
        chunks : int or dict, optional
            If not None, passed to xarray so that the full dataset is loaded
            and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        if engine is None:
            engine = self.engine

        if chunks is None:
            chunks = self.chunks

        # Check file exists and can be written to
        if os.access(self.data_name, os.W_OK):
            self._full_ds = load_ds(self.data_name,
                                    engine=engine,
                                    chunks=chunks)

        # Do nothing if file does not exist at all
        elif not os.path.isfile(self.data_name):  # pragma: no cover
            pass

        # Catch read-only errors etc.
        else:
            raise OSError("The file '{}' exists but cannot be written "
                          "to".format(self.data_name))

    @property
    def full_ds(self):
        """Dataset containing all saved runs.
        """
        if self._full_ds is None:
            self.load_full_ds()
        return self._full_ds

    def save_full_ds(self, new_full_ds=None, engine=None):
        """Save `full_ds` onto disk.

        Parameters
        ----------
        new_full_ds : xarray.Dataset, optional
            Save this dataset as the new full dataset, else use the current
            full datset.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        if engine is None:
            engine = self.engine

        if new_full_ds is not None:
            if os.path.exists(self.data_name):
                os.remove(self.data_name)
            self._full_ds = new_full_ds

        save_ds(self._full_ds, self.data_name, engine=engine)

    def delete_ds(self, backup=False):
        """Delete the on-disk dataset, optionally backing it up first.
        """
        if backup:
            import datetime
            ts = '{:%Y%m%d-%H%M%S}'.format(datetime.datetime.now())
            shutil.copy(self.data_name, self.data_name + '.BAK-{}'.format(ts))

        os.remove(self.data_name)

    def add_ds(self, new_ds,
               sync=True,
               overwrite=None,
               chunks=None,
               engine=None):
        """Merge a new dataset into the in-memory full dataset.

        Parameters
        ----------
        new_ds : xr.Dataset or xr.DataArray
            Data to be merged into the full dataset.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional
            How to combine data from the new run into the current full_ds:

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite conflicting current data with
              that from the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : int or dict, optional
            If not None, passed to xarray so that the full dataset is loaded
            and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        """
        if isinstance(new_ds, xr.DataArray):
            new_ds = new_ds.to_dataset()

        # get default chunks if not overidden
        if chunks is None:
            chunks = self.chunks
        if chunks is not None:
            new_ds = new_ds.chunk(chunks)

        # only sync with disk if data name present
        sync_with_disk = sync and self.data_name is not None
        if sync_with_disk:
            self.load_full_ds(chunks=chunks, engine=engine)

        if self._full_ds is None:
            # No full ds yet, deep copy to maintain distinction between
            #   'full_ds' and 'last_ds'.
            new_full_ds = new_ds.copy(deep=True)

        else:
            # Overwrite with new data
            if overwrite is True:
                new_full_ds = new_ds.combine_first(self._full_ds)
            # Overwrite nothing
            elif overwrite is False:
                new_full_ds = self._full_ds.combine_first(new_ds)
            # Merge, raising error if the two datasets conflict
            else:
                new_full_ds = self._full_ds.merge(
                    new_ds, compat='no_conflicts', inplace=False)

        if sync_with_disk:
            self.save_full_ds(new_full_ds, engine=engine)
        else:
            self._full_ds = new_full_ds

    def harvest_combos(self, combos, *,
                       sync=True,
                       overwrite=None,
                       chunks=None,
                       engine=None,
                       **runner_settings):
        """Run combos, automatically merging into an on-disk dataset.

        Parameters
        ---------
        combos : mapping_like
            The combos to run.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite conflicting current data with
              that from the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : bool, optional
            If not None, passed passed to xarray so that the full dataset is
            loaded and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        runner_settings
            Supplied to :func:`~xyzpy.combo_runner`.
        """

        ds = self.runner.run_combos(combos, **runner_settings)
        self.add_ds(ds, sync=sync, overwrite=overwrite,
                    chunks=chunks, engine=engine)

    def harvest_cases(self, cases, *,
                      sync=True,
                      overwrite=None,
                      chunks=None,
                      engine=None,
                      **runner_settings):
        """Run cases, automatically merging into an on-disk dataset.

        Parameters
        ---------
        cases : list of dict or tuple
            The cases to run.
        sync : bool, optional
            If True (default), load and save the disk dataset before
            and after merging in the new data.
        overwrite : {None, False, True}, optional
            What to do regarding clashes with old data:

            - ``None`` (default): attempt the merge and only raise if
              data conflicts.
            - ``True``: overwrite conflicting current data with
              that from the new dataset.
            - ``False``: drop any conflicting data from the new dataset.

        chunks : bool, optional
            If not None, passed passed to xarray so that the full dataset is
            loaded and merged into with on-disk dask arrays.
        engine : str, optional
            Engine to use to save and load datasets.
        runner_settings
            Supplied to :func:`~xyzpy.case_runner`.
        """
        ds = self.runner.run_cases(cases, **runner_settings)
        self.add_ds(ds, sync=sync, overwrite=overwrite, chunks=chunks,
                    engine=engine)

    def Crop(self,
             name=None,
             parent_dir=None,
             save_fn=None,
             batchsize=None,
             num_batches=None):
        """Return a Crop instance with this Harvester, from which `fn`
        will be set, and then combos can be sown, grown, and reaped into the
        ``Harvester.full_ds``. See :class:`~xyzpy.Crop`.

        Returns
        -------
        Crop
        """
        return Crop(harvester=self,
                    name=name,
                    parent_dir=parent_dir,
                    save_fn=save_fn,
                    batchsize=batchsize,
                    num_batches=num_batches)

    def __repr__(self):
        string = ("<xyzpy.Harvester>\n"
                  "Runner: {self.runner}"
                  "Sync file -->\n"
                  "    {self.data_name}    [{self.engine}]")

        return string.format(self=self)



import os
import shutil
from itertools import chain
from time import sleep
from glob import glob
import warnings
import pickle
import copy
import math

try:
    import joblib
except ImportError:  # pragma: no cover
    pass

try:
    import cloudpickle
except ImportError:  # pragma: no cover
    pass

from ..utils import _get_fn_name, prod, progbar
from .prepare import _parse_combos, _parse_constants, _parse_attrs
from .combo_runner import _combo_runner, combo_runner_to_ds


BTCH_NM = "xyz-batch-{}.jbdmp"
RSLT_NM = "xyz-result-{}.jbdmp"
FNCT_NM = "xyz-function.clpkl"
INFO_NM = "xyz-settings.jbdmp"


class XYZError(Exception):
    pass


# --------------------------------- parsing --------------------------------- #

def parse_crop_details(fn, crop_name, crop_parent):
    """Work out how to structure the sowed data.

    Parameters
    ----------
    fn : callable, optional
        Function to infer name crop_name from, if not given.
    crop_name : str, optional
        Specific name to give this set of runs.
    crop_parent : str, optional
        Specific directory to put the ".xyz-{crop_name}/" folder in
        with all the cases and results.

    Returns
    -------
    crop_location : str
        Full path to the crop-folder.
    crop_name : str
        Name of the crop.
    crop_parent : str
        Parent folder of the crop.
    """
    if crop_name is None:
        if fn is None:
            raise ValueError("Either `fn` or `crop_name` must be give.")
        crop_name = _get_fn_name(fn)

    crop_parent = crop_parent if crop_parent is not None else os.getcwd()
    crop_location = os.path.join(crop_parent, ".xyz-{}".format(crop_name))

    return crop_location, crop_name, crop_parent


def parse_fn_runner_harvester(fn, runner, harvester):
    """
    """
    if harvester is not None:
        if (runner is not None) or (fn is not None):
            warnings.warn(
                "If `harvester` is set `runner` and `fn` are ignored.")
        harvester = harvester
        runner = harvester.runner
        fn = harvester.runner.fn
    elif runner is not None:
        if fn is not None:
            warnings.warn("If `runner` is set `fn` is ignored.")
        harvester = None
        runner = runner
        fn = runner.fn
    else:
        harvester = None
        runner = None
        fn = fn
    return fn, runner, harvester


class Crop(object):
    """Encapsulates all the details describing a single 'crop', that is,
    its location, name, and batch size/number. Also allows tracking of
    crop's progress, and experimentally, automatic submission of
    workers to grid engine to complete un-grown cases. Can also be instantiated
    directly from a :class:`~xyzpy.Runner` or :class:`~xyzpy.Harvester`
    instance.

    Parameters
    ----------
    fn : callable, optional
        Target function - Crop `name` will be inferred from this if
        not given explicitly. If given, `Sower` will also default
        to saving a version of `fn` to disk for `batch.grow` to use.
    name : str, optional
        Custom name for this set of runs - must be given if `fn`
        is not.
    parent_dir : str, optional
        If given, alternative directory to put the ".xyz-{name}/"
        folder in with all the cases and results.
    save_fn : bool, optional
        Whether to save the function to disk for `batch.grow` to use.
        Will default to True if `fn` is given.
    batchsize : int, optional
        How many cases to group into a single batch per worker.
        By default, batchsize=1. Cannot be specified if `num_batches`
        is.
    num_batches : int, optional
        How many total batches to aim for, cannot be specified if
        `batchsize` is.
    runner : xyzpy.Runner, optional
        A Runner instance, from which the `fn` can be inferred and
        which can also allow the Crop to reap itself straight to a
        dataset.
    harvester : xyzpy.Harvester, optional
        A Harvester instance, from which the `fn` can be inferred and
        which can also allow the Crop to reap itself straight to a
        on-disk dataset.
    autoload : bool, optional
        If True, check for the existence of a Crop written to disk
        with the same location, and if found, load it.

    See Also
    --------
    Runner.Crop, Harvester.Crop
    """

    def __init__(self, *,
                 fn=None,
                 name=None,
                 parent_dir=None,
                 save_fn=None,
                 batchsize=None,
                 num_batches=None,
                 runner=None,
                 harvester=None,
                 autoload=True):

        self._fn, self.runner, self.harvester = \
            parse_fn_runner_harvester(fn, runner, harvester)

        self.name = name
        self.parent_dir = parent_dir
        self.save_fn = save_fn
        self.batchsize = batchsize
        self.num_batches = num_batches

        # Work out the full directory for the crop
        self.location, self.name, self.parent_dir = \
            parse_crop_details(self._fn, self.name, self.parent_dir)

        # Save function so it can be automatically loaded with all deps?
        if (fn is None) and (save_fn is True):
            raise ValueError("Must specify a function for it to be saved!")
        self.save_fn = save_fn is not False

    # ------------------------------- methods ------------------------------- #

    def choose_batch_settings(self, combos):
        """Work out how to divide all cases into batches, i.e. ensure
        that ``batchsize * num_batches >= num_cases``.
        """
        n = prod(len(x) for _, x in combos)

        if (self.batchsize is not None) and (self.num_batches is not None):
            # Check that they are set correctly
            pos_tot = self.batchsize * self.num_batches
            if not (n <= pos_tot < n + self.batchsize):
                raise ValueError("`batchsize` and `num_batches` cannot both"
                                 "be specified if they do not not multiply"
                                 "to the correct number of total cases.")

        # Decide based on batchsize
        elif self.num_batches is None:
            if self.batchsize is None:
                self.batchsize = 1

            if not isinstance(self.batchsize, int):
                raise TypeError("`batchsize` must be an integer.")
            if self.batchsize < 1:
                raise ValueError("`batchsize` must be >= 1.")

            self.num_batches = math.ceil(n / self.batchsize)
            self._batch_remainder = 0

        # Decide based on num_batches:
        else:
            if not isinstance(self.num_batches, int):
                raise TypeError("`num_batches` must be an integer.")
            if self.num_batches < 1:
                raise ValueError("`num_batches` must be >= 1.")

            self.batchsize, self._batch_remainder = divmod(n, self.num_batches)

    def ensure_dirs_exists(self):
        """Make sure the directory structure for this crop exists.
        """
        os.makedirs(os.path.join(self.location, "batches"), exist_ok=True)
        os.makedirs(os.path.join(self.location, "results"), exist_ok=True)

    def save_info(self, combos):
        """Save information about the sowed cases.
        """
        # If saving Harvester or Runner, strip out function information so
        #   as just to use pickle.
        if self.harvester is not None:
            harvester_copy = copy.deepcopy(self.harvester)
            harvester_copy.runner.fn = None
            hrvstr_pkl = pickle.dumps(harvester_copy)
            runner_pkl = None
        elif self.runner is not None:
            hrvstr_pkl = None
            runner_copy = copy.deepcopy(self.runner)
            runner_copy.fn = None
            runner_pkl = pickle.dumps(runner_copy)
        else:
            hrvstr_pkl = None
            runner_pkl = None

        joblib.dump({
            'combos': combos,
            'batchsize': self.batchsize,
            'num_batches': self.num_batches,
            'harvester': hrvstr_pkl,
            'runner': runner_pkl,
        }, os.path.join(self.location, INFO_NM))

    def load_info(self):
        """Load information about the saved cases.
        """
        settings = joblib.load(os.path.join(self.location, INFO_NM))
        self.batchsize = settings['batchsize']
        self.num_batches = settings['num_batches']

        hrvstr_pkl = settings['harvester']
        harvester = None if hrvstr_pkl is None else pickle.loads(hrvstr_pkl)
        runner_pkl = settings['runner']
        runner = None if runner_pkl is None else pickle.loads(runner_pkl)

        self._fn, self.runner, self.harvester = \
            parse_fn_runner_harvester(None, runner, harvester)

    def save_function_to_disk(self):
        """Save the base function to disk using cloudpickle
        """
        import cloudpickle

        joblib.dump(cloudpickle.dumps(self._fn),
                    os.path.join(self.location, FNCT_NM))

    def load_function(self):
        """Load the saved function from disk, and try to re-insert it back into
        Harvester or Runner if present.
        """
        import cloudpickle

        self._fn = cloudpickle.loads(joblib.load(
            os.path.join(self.location, FNCT_NM)))

        if self.harvester is not None:
            if self.harvester.runner.fn is None:
                self.harvester.runner.fn = self._fn
            else:
                # TODO: check equality?
                raise XYZError("Trying to load this Crop's function, {}, from "
                               "disk but its Harvester already has a function "
                               "set: {}.".format(self._fn,
                                                 self.harvester.runner.fn))
        elif self.runner is not None:
            if self.runner.fn is None:
                self.runner.fn = self._fn
            else:
                # TODO: check equality?
                raise XYZError("Trying to load this Crop's function, {}, from "
                               "disk but its Runner already has a function "
                               "set: {}.".format(self._fn,
                                                 self.runner.fn))

    def prepare(self, combos):
        """Write information about this crop and the supplied combos to disk.
        Typically done at start of sow, not when Crop instantiated.
        """
        self.ensure_dirs_exists()
        if self.save_fn:
            self.save_function_to_disk()
        self.save_info(combos)

    def is_prepared(self):
        """Check whether this crop has been written to disk.
        """
        return os.path.exists(os.path.join(self.location, INFO_NM))

    def calc_progress(self):
        """Calculate how much progressed has been made in growing the cases.
        """
        if self.is_prepared():
            self.load_info()
            self._num_sown_batches = len(glob(
                os.path.join(self.location, "batches", BTCH_NM.format("*"))))
            self._num_results = len(glob(
                os.path.join(self.location, "results", RSLT_NM.format("*"))))
        else:
            self._num_sown_batches = -1
            self._num_results = -1

    def is_ready_to_reap(self):
        self.calc_progress()
        return (
            self._num_results > 0 and
            (self._num_results == self.num_sown_batches)
        )

    def missing_results(self):
        """
        """
        self.calc_progress()

        def no_result_exists(x):
            return not os.path.isfile(
                os.path.join(self.location, "results", RSLT_NM.format(x)))

        return tuple(filter(no_result_exists, range(1, self.num_batches + 1)))

    def delete_all(self):
        # delete everything
        shutil.rmtree(self.location)

    def __str__(self):
        # Location and name, underlined
        if not os.path.exists(self.location):
            return self.location + "\n * Not yet sown, or already reaped * \n"

        loc_len = len(self.location)
        name_len = len(self.name)

        self.calc_progress()
        percentage = 100 * self._num_results / self.num_batches

        # Progress bar
        total_bars = 20
        bars = int(percentage * total_bars / 100)

        return ("\n"
                "{location}\n"
                "{under_crop_dir}{under_crop_name}\n"
                "{num_results} / {total} batches of size {bsz} completed\n"
                "[{done_bars}{not_done_spaces}] : {percentage:.1f}%"
                "\n").format(
            location=self.location,
            under_crop_dir="-" * (loc_len - name_len),
            under_crop_name="=" * name_len,
            num_results=self._num_results,
            total=self.num_batches,
            bsz=self.batchsize,
            done_bars="#" * bars,
            not_done_spaces=" " * (total_bars - bars),
            percentage=percentage,
        )

    def __repr__(self):
        if not os.path.exists(self.location):
            progress = "*reaped or unsown*"
        else:
            self.calc_progress()
            progress = "{}/{}".format(self._num_results, self.num_batches)

        msg = "<Crop(name='{}', progress={}, batchsize={})>"
        return msg.format(self.name, progress, self.batchsize)

    def sow_combos(self, combos, constants=None, verbosity=1):
        """Sow to disk.
        """
        combos = _parse_combos(combos)
        constants = _parse_constants(constants)

        if self.runner is not None:
            constants = {**self.runner._constants, **constants}
            constants = {**self.runner._resources, **constants}

        # Sort to ensure order remains same for reaping results
        #   (don't want to hash kwargs)
        combos = sorted(combos, key=lambda x: x[0])

        with Sower(self, combos) as sow_fn:
            _combo_runner(fn=sow_fn, combos=combos, constants=constants,
                          verbosity=verbosity)

    def grow(self, batch_ids, **combo_runner_opts):
        """Grow specific batch numbers using this process.
        """
        if isinstance(batch_ids, int):
            batch_ids = (batch_ids,)

        _combo_runner(grow, combos=(('batch_number', batch_ids),),
                      constants={'verbosity': 0, 'crop': self},
                      **combo_runner_opts)

    def grow_missing(self, **combo_runner_opts):
        """Grow any missing results using this process.
        """
        self.grow(batch_ids=self.missing_results(), **combo_runner_opts)

    def reap_combos(self, wait=False, clean_up=True):
        """Reap already sown and grow results from specified crop.

        Parameters
        ----------
        wait : bool, optional
            Whether to wait for results to appear. If false (default) all
            results need to be in place before the reap.

        Returns
        -------
        results : nested tuple
            'N-dimensional' tuple containing the results.
        """
        if not (wait or self.is_ready_to_reap()):
            raise XYZError("This crop is not ready to reap yet - results are"
                           " missing.")

        # Load same combinations as cases saved with
        settings = joblib.load(os.path.join(self.location, INFO_NM))

        with Reaper(self, num_batches=settings['num_batches'],
                    wait=wait) as reap_fn:

            results = _combo_runner(fn=reap_fn,
                                    combos=settings['combos'],
                                    constants={})

        if clean_up:
            self.delete_all()

        return results

    def reap_combos_to_ds(self,
                          var_names=None,
                          var_dims=None,
                          var_coords=None,
                          constants=None,
                          attrs=None,
                          parse=True,
                          wait=False,
                          clean_up=True):
        """Reap a function over sowed combinations and output to a Dataset.

        Parameters
        ----------
        var_names : str, sequence of strings, or None
            Variable name(s) of the output(s) of `fn`, set to None if
            fn outputs data already labelled in a Dataset or DataArray.
        var_dims : sequence of either strings or string sequences, optional
            'Internal' names of dimensions for each variable, the values for
            each dimension should be contained as a mapping in either
            `var_coords` (not needed by `fn`) or `constants` (needed by `fn`).
        var_coords : mapping, optional
            Mapping of extra coords the output variables may depend on.
        constants : mapping, optional
            Arguments to `fn` which are not iterated over, these will be
            recorded either as attributes or coordinates if they are named
            in `var_dims`.
        resources : mapping, optional
            Like `constants` but they will not be recorded.
        attrs : mapping, optional
            Any extra attributes to store.
        wait : bool, optional
            Whether to wait for results to appear. If false (default) all
            results need to be in place before the reap.

        Returns
        -------
        xarray.Dataset
            Multidimensional labelled dataset contatining all the results.
        """
        if not (wait or self.is_ready_to_reap()):
            raise XYZError(
                "This crop is not ready to reap yet - results are missing.")

        # Load exact same combinations as cases saved with
        settings = joblib.load(os.path.join(self.location, INFO_NM))

        if parse:
            constants = _parse_constants(constants)
            attrs = _parse_attrs(attrs)

        with Reaper(self, num_batches=settings['num_batches'],
                    wait=wait) as reap_fn:

            # Move constants into attrs, so as not to pass them to the Reaper
            #   when if fact they were meant for the original function.
            ds = combo_runner_to_ds(fn=reap_fn,
                                    combos=settings['combos'],
                                    var_names=var_names,
                                    var_dims=var_dims,
                                    var_coords=var_coords,
                                    constants={},
                                    resources={},
                                    attrs={**attrs, **constants},
                                    parse=parse)

        if clean_up:
            self.delete_all()

        return ds

    def reap_runner(self, runner, wait=False, clean_up=True):
        """Reap a Crop over sowed combos and save to a dataset defined by a
        Runner.
        """
        # Can ignore `Runner.resources` as they play no part in desecribing the
        #   output, though they should be supplied to sow and thus grow.
        ds = self.reap_combos_to_ds(
            var_names=runner._var_names,
            var_dims=runner._var_dims,
            var_coords=runner._var_coords,
            constants=runner._constants,
            attrs=runner._attrs,
            parse=False,
            wait=wait,
            clean_up=clean_up)
        runner.last_ds = ds
        return ds

    def reap_harvest(self, harvester, wait=False, sync=True, overwrite=None,
                     clean_up=True):
        """
        """
        if harvester is None:
            raise ValueError("Cannot reap and harvest if no Harvester is set.")

        ds = self.reap_runner(harvester.runner, wait=wait, clean_up=clean_up)
        self.harvester.add_ds(ds, sync=sync, overwrite=overwrite)
        return ds

    def reap(self, wait=False, sync=True, overwrite=None, clean_up=True):
        """Reap sown and grown combos from disk. Return a dataset if a runner
        or harvester is set, otherwise, the raw nested tuple.

        Parameters
        ----------
        wait : bool, optional
            Whether to wait for results to appear. If false (default) all
            results need to be in place before the reap.
        sync : bool, optional
            Immediately sync the new dataset with the on-disk full dataset
            if a harvester is used.
        overwrite : bool, optional
            How to compare data when syncing to on-disk dataset.
            If ``None``, (default) merge as long as no conflicts.
            ``True``: overwrite with the new data. ``False``, discard any
            new conflicting data.
        clean_up : bool, optional
            Whether to delete all the batch files once the results have been
            gathered.

        Returns
        -------
        nested tuple or xarray.Dataset
        """
        if self.harvester is not None:
            return self.reap_harvest(self.harvester, clean_up=clean_up,
                                     wait=wait, sync=sync, overwrite=overwrite)
        elif self.runner is not None:
            return self.reap_runner(self.runner, wait=wait, clean_up=clean_up)
        else:
            return self.reap_combos(wait=wait, clean_up=clean_up)

    def check_bad(self, delete_bad=True):
        """Check that the result dumps are not bad -> sometimes length does not
        match the batch. Optionally delete these so that they can be re-grown.

        Parameters
        ----------
        delete_bad : bool
            Delete bad results as they are come across.

        Returns
        -------
        bad_ids : tuple
            The bad batch numbers.
        """
        # XXX: work out why this is needed sometimes on network filesystems.
        result_files = glob(
            os.path.join(self.location, "results", RSLT_NM.format("*")))

        bad_ids = []

        for result_file in result_files:
            # load corresponding batch file to check length.
            result_num = os.path.split(
                result_file)[-1].strip("xyz-result-").strip(".jbdmp")
            batch_file = os.path.join(
                self.location, "batches", BTCH_NM.format(result_num))

            batch = joblib.load(batch_file)

            try:
                result = joblib.load(result_file)
                unloadable = False
            except Exception as e:
                unloadable = True
                err = e

            if unloadable or (len(result) != len(batch)):
                msg = "result {} is bad".format(result_file)
                msg += "." if not delete_bad else " - deleting it."
                msg += " Error was: {}".format(err) if unloadable else ""
                print(msg)

                if delete_bad:
                    os.remove(result_file)

                bad_ids.append(result_num)

        return tuple(bad_ids)

    #  ----------------------------- properties ----------------------------- #

    def _get_fn(self):
        return self._fn

    def _set_fn(self, fn):
        if self.save_fn is None and fn is not None:
            self.save_fn = True
        self._fn = fn

    def _del_fn(self):
        self._fn = None
        self.save_fn = False

    fn = property(_get_fn, _set_fn, _del_fn,
                  "Function to save with the Crop for automatic loading and "
                  "running. Default crop name will be inferred from this if"
                  "not given explicitly as well.")

    @property
    def num_sown_batches(self):
        """Total number of batches to be run/grown.
        """
        self.calc_progress()
        return self._num_sown_batches

    @property
    def num_results(self):
        self.calc_progress()
        return self._num_results


class Sower(object):
    """Class for sowing a 'crop' of batched combos to then 'grow' (on any
    number of workers sharing the filesystem) and then reap.
    """

    def __init__(self, crop, combos):
        """
        Parameters
        ----------
            crop : xyzpy.batch.Crop instance
                Description of where and how to store the cases and results.
            combos : mapping_like
                Description of combinations from which to sow cases from.

        """
        self.combos = combos
        self.crop = crop
        self.crop.choose_batch_settings(combos)
        # Internal:
        self._batch_cases = []  # collects cases to be written in single batch
        self._counter = 0  # counts how many cases are in batch so far
        self._batch_counter = 0  # counts how many batches have been written

    def save_batch(self):
        """Save the current batch of cases to disk using joblib.dump
         and start the next batch.
        """
        self._batch_counter += 1
        joblib.dump(self._batch_cases, os.path.join(
            self.crop.location, "batches", BTCH_NM.format(self._batch_counter))
        )
        self._batch_cases = []
        self._counter = 0

    # Context manager #

    def __enter__(self):
        self.crop.prepare(self.combos)
        return self

    def __call__(self, **kwargs):
        self._batch_cases.append(kwargs)
        self._counter += 1

        # when the number of cases doesn't divide the number of batches we
        #     distribute the remainder among the first crops.
        extra_batch = self._batch_counter < self.crop._batch_remainder

        if self._counter == self.crop.batchsize + int(extra_batch):
            self.save_batch()

    def __exit__(self, exception_type, exception_value, traceback):
        # Make sure any overfill also saved
        if self._batch_cases:
            self.save_batch()


def grow(batch_number, crop=None, fn=None, check_mpi=True,
         verbosity=1, debugging=False):
    """Automatically process a batch of cases into results. Should be run in an
    ".xyz-{fn_name}" folder.

    Parameters
    ----------
    batch_number : int
        Which batch to 'grow' into a set of results.
    crop : xyzpy.batch.Crop instance
        Description of where and how to store the cases and results.
    fn : callable, optional
        If specified, the function used to generate the results, otherwise
        the function will be loaded from disk.
    check_mpi : bool, optional
        Whether to check if the process is rank 0 and only save results if
        so - allows mpi functions to be simply used. Defaults to true,
        this should only be turned off if e.g. a pool of workers is being
        used to run different ``grow`` instances.
    verbosity : {0, 1, 2}, optional
        How much information to show.
    debugging : bool, optional
        Set logging level to DEBUG.
    """
    if debugging:
        import logging
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)

    if crop is None:
        if os.path.relpath('.', '..')[:5] != ".xyz-":
            raise XYZError("`grow` should be run in a "
                           "\"{crop_parent}/.xyz-{crop_name}\" folder, else "
                           "`crop_parent` and `crop_name` (or `fn`) should be "
                           "specified.")
        crop_location = os.getcwd()
    else:
        crop_location = crop.location

    # load function
    if fn is None:
        fn = cloudpickle.loads(
            joblib.load(os.path.join(crop_location, FNCT_NM)))

    # load cases to evaluate
    cases = joblib.load(
        os.path.join(crop_location, "batches", BTCH_NM.format(batch_number)))

    if len(cases) == 0:
        raise ValueError("Something has gone wrong with the loading of "
                         "batch {} ".format(BTCH_NM.format(batch_number)) +
                         "for the crop at {}.".format(crop.location))

    # maybe want to run grow as mpiexec (i.e. `fn` itself in parallel),
    # so only save and delete on rank 0
    if check_mpi and 'OMPI_COMM_WORLD_RANK' in os.environ:  # pragma: no cover
        rank = int(os.environ['OMPI_COMM_WORLD_RANK'])
    elif check_mpi and 'PMI_RANK' in os.environ:  # pragma: no cover
        rank = int(os.environ['PMI_RANK'])
    else:
        rank = 0

    if rank == 0:
        results = []
        for i in progbar(range(len(cases)), disable=verbosity <= 0,
                         desc="Batch: {}".format(batch_number)):
            results.append(fn(**cases[i]))

        if len(results) != len(cases):
            raise ValueError("Something has gone wrong with processing "
                             "batch {} ".format(BTCH_NM.format(batch_number)) +
                             "for the crop at {}.".format(crop.location))

        # save to results
        joblib.dump(tuple(results), os.path.join(
            crop_location, "results", RSLT_NM.format(batch_number)))
    else:
        for i in range(len(cases)):
            fn(**cases[i])


# --------------------------------------------------------------------------- #
#                              Gathering results                              #
# --------------------------------------------------------------------------- #

class Reaper(object):
    """Class that acts as a stateful function to retrieve already sown and
    grow results.
    """

    def __init__(self, crop, num_batches, wait=False):
        """Class for retrieving the batched, flat, 'grown' results.

        Parameters
        ----------
            crop : xyzpy.batch.Crop instance
                Description of where and how to store the cases and results.
        """
        self.crop = crop

        files = (os.path.join(self.crop.location, "results", RSLT_NM.format(i))
                 for i in range(1, num_batches + 1))

        def _load(x):
            res = joblib.load(x)
            if (res is None) or len(res) == 0:
                raise ValueError("Something not right: result {} contains "
                                 "no data upon joblib.load".format(x))
            return res

        def wait_to_load(x):
            while not os.path.exists(x):
                sleep(0.2)

            if os.path.isfile(x):
                return _load(x)
            else:
                raise ValueError("{} is not a file.".format(x))

        self.results = chain.from_iterable(map(
            wait_to_load if wait else _load, files))

    def __enter__(self):
        return self

    def __call__(self, **kwargs):
        return next(self.results)

    def __exit__(self, exception_type, exception_value, traceback):
        # Check everything gone acccording to plan
        if tuple(self.results):
            raise XYZError("Not all results reaped!")


# --------------------------------------------------------------------------- #
#                     Automatic Batch Submission Scripts                      #
# --------------------------------------------------------------------------- #

_BASE_QSUB_SCRIPT = """#!/bin/bash -l
#$ -S /bin/bash
#$ -l h_rt={hours}:{minutes}:{seconds}
#$ -l mem={gigabytes}G
#$ -l tmpfs={temp_gigabytes}G
#$ -N {name}
mkdir -p {output_directory}
#$ -wd {output_directory}
#$ -pe {pe} {num_procs}
#$ -t {run_start}-{run_stop}
cd {working_directory}
export OMP_NUM_THREADS={num_threads}
tmpfile=$(mktemp .xyzpy-qsub.XXXXXXXX)
cat <<EOF > $tmpfile
from xyzpy.gen.batch import grow, Crop
crop = Crop(name='{name}')
"""

_QSUB_GROW_ALL_SCRIPT = """grow($SGE_TASK_ID, crop=crop, debugging={debugging})
"""

_QSUB_GROW_PARTIAL_SCRIPT = """batch_ids = {batch_ids}
grow(batch_ids[$SGE_TASK_ID - 1], crop=crop, debugging={debugging})
"""

_BASE_QSUB_SCRIPT_END = """EOF
{launcher} $tmpfile
rm $tmpfile
"""


def gen_qsub_script(crop, batch_ids=None, *,
                    hours=None,
                    minutes=None,
                    seconds=None,
                    gigabytes=2,
                    num_procs=1,
                    launcher='python',
                    mpi=False,
                    temp_gigabytes=1,
                    output_directory=None,
                    debugging=False):
    """Generate a qsub script to grow a Crop.

    Parameters
    ----------
    crop : Crop
        The crop to grow.
    batch_ids : int or tuple[int]
        Which batch numbers to grow, defaults to all missing batches.
    hours : int
        How many hours to request, default=0.
    minutes : int, optional
        How many minutes to request, default=20.
    seconds : int, optional
        How many seconds to request, default=0.
    gigabytes : int, optional
        How much memory to request, default: 2.
    num_procs : int, optional
        How many processes to request (threaded cores or MPI), default: 1.
    launcher : str, optional
        How to launch the script, default: ``'python'``. But could for example
        be ``'mpiexec python'`` for a MPI program.
    mpi : bool, optional
        Request MPI processes not threaded processes.
    temp_gigabytes : int, optional
        How much temporary on-disk memory.
    output_directory : str, optional
        What directory to write output to. Defaults to "$HOME/scratch/output".
    debugging : bool, optional
        Set the python log level to debugging.

    Returns
    -------
    str
    """
    if hours is minutes is seconds is None:
        hours, minutes, seconds = 1, 0, 0
    else:
        hours = 0 if hours is None else int(hours)
        minutes = 0 if minutes is None else int(minutes)
        seconds = 0 if seconds is None else int(seconds)

    if output_directory is None:
        from os.path import expanduser
        home = expanduser("~")
        output_directory = os.path.join(home, 'Scratch', 'output')

    crop.calc_progress()

    opts = {
        'hours': hours,
        'minutes': minutes,
        'seconds': seconds,
        'gigabytes': gigabytes,
        'name': crop.name,
        'num_procs': num_procs,
        'num_threads': 1 if mpi else num_procs,
        'run_start': 1,
        'launcher': launcher,
        'pe': 'mpi' if mpi else 'smp',
        'temp_gigabytes': temp_gigabytes,
        'output_directory': output_directory,
        'working_directory': crop.parent_dir,
        'debugging': debugging,
    }

    script = _BASE_QSUB_SCRIPT

    # grow specific ids
    if batch_ids is not None:
        script += _QSUB_GROW_PARTIAL_SCRIPT
        batch_ids = tuple(batch_ids)
        opts['run_stop'] = len(batch_ids)
        opts['batch_ids'] = batch_ids

    # grow all ids
    elif crop.num_results == 0:
        script += _QSUB_GROW_ALL_SCRIPT
        opts['run_stop'] = crop.num_batches

    # grow missing ids only
    else:
        script += _QSUB_GROW_PARTIAL_SCRIPT
        batch_ids = crop.missing_results()
        opts['run_stop'] = len(batch_ids)
        opts['batch_ids'] = batch_ids

    script += _BASE_QSUB_SCRIPT_END

    return script.format(**opts)


def qsub_grow(crop, batch_ids=None, *,
              hours=None,
              minutes=None,
              seconds=None,
              gigabytes=2,
              num_procs=1,
              launcher='python',
              mpi=False,
              temp_gigabytes=1,
              output_directory=None,
              debugging=False):  # pragma: no cover
    """Automagically submit SGE jobs to grow all missing results.

    Parameters
    ----------
    crop : Crop
        The crop to grow.
    batch_ids : int or tuple[int]
        Which batch numbers to grow, defaults to all missing batches.
    hours : int
        How many hours to request, default=0.
    minutes : int, optional
        How many minutes to request, default=20.
    seconds : int, optional
        How many seconds to request, default=0.
    gigabytes : int, optional
        How much memory to request, default: 2.
    num_procs : int, optional
        How many processes to request (threaded cores or MPI), default: 1.
    launcher : str, optional
        How to launch the script, default: ``'python'``. But could for example
        be ``'mpiexec python'`` for a MPI program.
    mpi : bool, optional
        Request MPI processes not threaded processes.
    temp_gigabytes : int, optional
        How much temporary on-disk memory.
    output_directory : str, optional
        What directory to write output to. Defaults to "$HOME/scratch/output".
    debugging : bool, optional
        Set the python log level to debugging.
    """
    if crop.is_ready_to_reap():
        print("Crop ready to reap: nothing to submit.")
        return

    import subprocess

    script = gen_qsub_script(crop,
                             batch_ids=batch_ids,
                             hours=hours,
                             minutes=minutes,
                             seconds=seconds,
                             gigabytes=gigabytes,
                             temp_gigabytes=temp_gigabytes,
                             output_directory=output_directory,
                             num_procs=num_procs,
                             launcher=launcher,
                             mpi=mpi,
                             debugging=debugging)

    script_file = os.path.join(crop.location, "__qsub_script__.sh")

    with open(script_file, mode='w') as f:
        f.write(script)

    subprocess.run(['qsub', script_file])

    os.remove(script_file)


Crop.gen_qsub_script = gen_qsub_script
Crop.qsub_grow = qsub_grow



import collections

_MPL_MARKER_DICT = collections.OrderedDict([
    ('circle', 'o'),
    ('x', 'x'),
    ('diamond', 'D'),
    ('plus', '+'),
    ('square', 's'),
    ('point', '.'),
    ('triangle_up', '^'),
    ('tri_left', '3'),
    ('triangle_right', '>'),
    ('thin_diamond', 'd'),
    ('star', '*'),
    ('triangle_down', 'v'),
    ('vline', '|'),
    ('tri_down', '1'),
    ('pentagon', 'p'),
    ('pixel', ','),
    ('tri_up', '2'),
    ('triangle_left', '<'),
    ('hexagon1', 'h'),
    ('tri_right', '4'),
    ('tickleft', 0),
    ('tickup', 2),
    ('tickdown', 3),
    ('caretleft', 4),
    ('hline', '_'),
    ('caretright', 5),
    ('hexagon2', 'H'),
    ('tickright', 1),
    ('caretup', 6),
    ('octagon', '8'),
    ('caretdown', 7)])
_MPL_MARKERS = [*_MPL_MARKER_DICT.values()]

_BOKEH_MARKER_DICT = collections.OrderedDict([
    ('circle', 'circle'),
    ('x', 'x'),
    ('diamond', 'diamond'),
    ('plus', 'cross'),
    ('square', 'square'),
    # ('point', '.'),
    ('triangle_up', 'triangle'),
    # ('tri_left', '3'),
    # ('triangle_right', '>'),
    # ('thin_diamond', 'd'),
    ('star', 'asterisk'),
    ('triangle_down', 'inverted_triangle'),
    # ('vline', '|'),
    # ('tri_down', '1'),
    # ('pentagon', 'p'),
    # ('pixel', ','),
    # ('tri_up', '2'),
    # ('triangle_left', '<'),
    # ('hexagon1', 'h'),
    # ('tri_right', '4'),
    # ('tickleft', 0),
    # ('tickup', 2),
    # ('tickdown', 3),
    # ('caretleft', 4),
    # ('hline', '_'),
    # ('caretright', 5),
    # ('hexagon2', 'H'),
    # ('tickright', 1),
    # ('caretup', 6),
    # ('octagon', '8'),
    # ('caretdown', 7)
    ('circle_cross', 'circle_cross'),
    ('square_x', 'square_x'),
    ('diamond_cross', 'diamond_cross'),
    ('circle_x', 'circle_x'),
    ('square_cross', 'square_cross')])
_BOKEH_MARKERS = [*_BOKEH_MARKER_DICT.values()]

_MARKERS = {'MATPLOTLIB': _MPL_MARKERS, 'BOKEH': _BOKEH_MARKERS}
_SINGLE_LINE_MARKER = {'MATPLOTLIB': ('.',), 'BOKEH': ('circle',)}



from setuptools import setup, find_packages
import versioneer


def readme():
    with open('README.rst') as f:
        import re
        long_desc = f.read()
        # strip out the raw html images
        long_desc = re.sub('\.\. raw::[\S\s]*?>\n\n', "", long_desc)
        return long_desc


setup(
    name='xyzpy',
    description='Easily generate large parameter space data',
    long_description=readme(),
    url='http://xyzpy.readthedocs.io',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Johnnie Gray',
    author_email="john.gray.14@ucl.ac.uk",
    license='MIT',
    packages=find_packages(exclude=['docs', 'test*']),
    install_requires=[
        'numpy>=1.10.0',
        'dask>=0.11.1',
        'xarray>=0.9.0',
        'h5py>=2.6.0',
        'h5netcdf>=0.2.2',
        'joblib>=0.10',
        'tqdm>=4.7.6',
        'matplotlib>=2.2.0',
        'bokeh>=0.12.3',
        'cytoolz>=0.8',
    ],
    extras_require={
        'tests': [
            'coverage',
            'pytest',
            'pytest-cov',
        ],
        'docs': [
            'sphinx',
            'sphinx_bootstrap_theme',
            'nbsphinx',
            'ipython',
        ]
    },
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



"""
Helper functions for generating color spectrums.
"""
import itertools
import matplotlib.cm as cm
from matplotlib.colors import BASE_COLORS, CSS4_COLORS, to_rgba
from .xyz_cmaps import _XYZ_CMAPS


def _COLORS_MATPLOTLIB_TO_BOKEH(cols):
    for col in cols:
        yield (int(255 * col[0]), int(255 * col[1]), int(255 * col[2]), col[3])


_COLOR_CONVERT_METHODS = {
    ('MATPLOTLIB', 'BOKEH'): _COLORS_MATPLOTLIB_TO_BOKEH,
}


def convert_colors_string_to_tuple(cols):
    for col in cols:
        if isinstance(col, str):
            if col in BASE_COLORS:
                yield to_rgba(BASE_COLORS[col])
            elif col in CSS4_COLORS:
                yield to_rgba(CSS4_COLORS[col])
            else:
                raise ValueError("Color specifier {} "
                                 "not recognized".format(col))
        else:
            yield to_rgba(col)


def convert_colors(cols, outformat, informat='MATPLOTLIB'):
    """Convert lists of colors between formats
    """
    # make sure cols are numeric tuples first
    cols = convert_colors_string_to_tuple(cols)

    if informat == outformat:
        return cols
    return _COLOR_CONVERT_METHODS[(informat, outformat)](cols)


def xyz_colormaps(name=None, reverse=False):
    """Custom-defined colormaps
    """
    try:
        import colorcet
        found_colorcet = True
    except (ImportError, KeyError):
        found_colorcet = False
    try:
        import cmocean
    except (ImportError, KeyError):
        pass

    # the default colormap
    if name is None:
        if found_colorcet:
            name = 'rainbow'
        else:
            name = 'xyz'

    # specify some aliases
    name = {
        'rainbow_pink': 'rainbow_bgyrm_35_85_c71',
        'rainbow_pink_r': 'rainbow_bgyrm_35_85_c71_r',
        'isolum_pink': 'isoluminant_cm_70_c39',
        'isolum_pink_r': 'isoluminant_cm_70_c39_r',
    }.get(name, name)

    # TODO: make this more general - not reliant on '_r' versions of cmaps
    if reverse:
        if name[-2:] != '_r':
            name = name + '_r'
        # 'un'-reverse
        else:
            name = name[:-2]

    # Custom xyzpy colormaps
    if name in _XYZ_CMAPS:
        return _XYZ_CMAPS[name]

    # special cases with name conflicts -->  prefer matplotlib
    if name not in {'inferno', 'coolwarm', 'blues'}:
        try:
            return colorcet.cm[name]
        except (NameError, KeyError):
            pass
        try:
            return getattr(cmocean.cm, name)
        except (NameError, AttributeError):
            pass

    # matplotlib colormaps
    return getattr(cm, name)


def get_default_sequential_cm(engine='MATPLOTLIB'):
    if engine == 'BOKEH':
        from bokeh.palettes import Category10_9
        return itertools.cycle(Category10_9)
    else:
        return itertools.cycle(rgb + (1.,) for rgb in cm.tab10.colors)



"""Functions for systematically evaluating a function over specific cases.
"""

import itertools
from cytoolz import concat
import numpy as np
import xarray as xr

from ..utils import progbar
from .prepare import (
    _parse_fn_args,
    _parse_cases,
    _parse_case_results,
    _parse_var_names,
    _parse_var_dims,
    _parse_var_coords,
    _parse_constants,
    _parse_resources
)


from .combo_runner import _combo_runner


class SingleArgFn:

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, kws, **kwargs):
        return self.fn(**kws, **kwargs)


def _case_runner(fn, fn_args, cases, constants,
                 split=False,
                 parallel=False,
                 num_workers=None,
                 pool=None,
                 verbosity=1):
    """Core case runner, i.e. without parsing of arguments.
    """
    # Turn the function into a single arg function to send to combo_runner
    sfn = SingleArgFn(fn)

    if isinstance(cases[0], dict):
        combos = (('kws', cases),)
    else:
        combos = (('kws', [dict(zip(fn_args, case)) for case in cases]),)

    return _combo_runner(sfn, combos,
                         constants=constants,
                         split=split,
                         parallel=parallel,
                         num_workers=num_workers,
                         pool=pool,
                         verbosity=verbosity)


def case_runner(fn, fn_args, cases,
                constants=None,
                split=False,
                parallel=False,
                pool=None,
                num_workers=None,
                verbosity=1):
    """Evaluate a function in many different configurations, optionally in
    parallel and or with live progress.

    Parameters
    ----------
    fn : callable
        Function with which to evalute cases with
    fn_args : tuple
        Names of case arguments that fn takes
    cases : tuple of tuple
        List settings that ``fn_args`` take.
    constants : dict, optional
        See :func:`~xyzpy.combo_runner`.
    split : bool, optional
        See :func:`~xyzpy.combo_runner`.
    parallel : bool, optional
        See :func:`~xyzpy.combo_runner`.
    pool : executor-like pool, optional
        See :func:`~xyzpy.combo_runner`.
    num_workers : int, optional
        See :func:`~xyzpy.combo_runner`.
    verbosity : {0, 1, 2}, optional
        See :func:`~xyzpy.combo_runner`.

    Returns
    -------
        results : list of fn output for each case
    """
    # Prepare fn_args and values
    fn_args = _parse_fn_args(fn_args)
    cases = _parse_cases(cases)
    constants = _parse_constants(constants)

    return _case_runner(fn, fn_args, cases,
                        constants=constants,
                        split=split,
                        parallel=parallel,
                        num_workers=num_workers,
                        pool=pool,
                        verbosity=verbosity)


def find_union_coords(cases):
    """Take a list of cases and find the union of coordinates
    with which to index all cases. Sort the coords if possible.
    """
    for x in zip(*cases):
        try:
            yield sorted(list(set(x)))
        except TypeError:  # unsortable
            yield list(set(x))


def all_missing_ds(coords, var_names, all_dims, var_types, attrs=None):
    """Make a dataset whose data is all missing.

    Parameters
    ----------
    coords : dict
        coordinates of dataset
    var_names : tuple
        names of each variable in dataset
    all_dims : tuple
        corresponding list of dimensions for each variable
    var_types : tuple
        corresponding list of types for each variable
    """
    # Blank dataset with appropirate coordinates
    ds = xr.Dataset(coords=coords, attrs=attrs)

    for v_name, v_dims, v_type in zip(var_names, all_dims, var_types):

        shape = tuple(ds[d].size for d in v_dims)

        if v_type == int or v_type == float:
            # Warn about upcasting int to float?
            nodata = np.tile(np.nan, shape)
        elif v_type == complex:
            nodata = np.tile(np.nan + np.nan * 1.0j, shape)
        else:
            nodata = np.tile(None, shape).astype(object)
        ds[v_name] = (v_dims, nodata)

    return ds


def _cases_to_ds(results, fn_args, cases, var_names, add_to_ds=None,
                 var_dims=None, var_coords=None, constants=None, attrs=None,
                 overwrite=False):
    """ Take a list of results and configurations that generate them and turn it
    into a `xarray.Dataset`.

    Parameters
    ----------
    results :
        List(s) of results of len(cases), e.g. generated by `case_runner`.
    fn_args :
        Arguments used in function that generated the results.
    cases :
        List of configurations used to generate results.
    var_names:
        Name(s) of output variables for a single result.
    var_dims :
        The list of named coordinates for each single result variable, i.e.
        coordinates not generated by the combo_runner.
    var_coords: dict of values for those coordinates if custom ones are
        desired.

    Returns
    -------
    ds: Dataset holding all results, with coordinates described by cases

    Notes
    -----
    1. Many data types have to be converted to object in order for the
        missing values to be represented by NaNs.
    """
    results = _parse_case_results(results, var_names)

    # check if Dataset already returned -> can just merge
    if isinstance(results[0][0], (xr.Dataset, xr.DataArray)):
        ds0 = (add_to_ds,) if add_to_ds is not None else ()
        return xr.merge([*ds0, *(r[0] for r in results)])

    if add_to_ds is not None:
        ds = add_to_ds
    else:
        # Find minimal covering set of coordinates for fn_args
        if isinstance(cases[0], dict):
            fn_args = tuple(cases[0].keys())
            cases = tuple(tuple(c[a] for a in fn_args) for c in cases)

        case_coords = dict(zip(fn_args, find_union_coords(cases)))

        # Create new, 'all missing' dataset if required
        ds = all_missing_ds(coords={**case_coords, **var_coords},
                            var_names=var_names, attrs=attrs,
                            all_dims=tuple(fn_args + var_dims[k]
                                           for k in var_names),
                            var_types=(np.asarray(x).dtype
                                       for x in results[0]))
        if constants:
            ds.attrs.update({k: v for k, v in constants.items()
                             if k not in ds.dims})

    # Go through cases, overwriting nan with results
    for res, cfg in zip(results, cases):

        cfg = [[c] for c in cfg]

        for vname, x in zip(var_names, res):
            if not overwrite:
                if not ds[vname].loc[dict(zip(fn_args, cfg))].isnull().all():
                    raise ValueError("Existing data and `overwrite` = False")
            try:
                len(x)
                ds[vname].loc[dict(zip(fn_args, cfg))] = np.asarray(x)
            except (TypeError, ValueError):
                ds[vname].loc[dict(zip(fn_args, cfg))] = x

    return ds


def case_runner_to_ds(fn, fn_args, cases, var_names,
                      var_dims=None,
                      var_coords=None,
                      constants=None,
                      resources=None,
                      attrs=None,
                      add_to_ds=None,
                      overwrite=False,
                      parse=True,
                      **case_runner_settings):
    """ Combination of `case_runner` and `_cases_to_ds`. Takes a function and
    list of argument configurations and produces a `xarray.Dataset`.

    Parameters
    ----------
    fn: function to evaluate
    fn_args: names of function args
    cases: list of function arg configurations
    var_names: list of names of single fn output
    var_dims: list of list of extra dims for each fn output
    var_coords: dictionary describing custom values of var_dims
    case_runner_settings: dict to supply to `case_runner`

    Returns
    -------
        ds: dataset with minimal covering coordinates and all cases
            evaluated.
    """
    if parse:
        fn_args = _parse_fn_args(fn_args)
        cases = _parse_cases(cases)
        constants = _parse_constants(constants)
        resources = _parse_resources(resources)
        var_names = _parse_var_names(var_names)
        var_dims = _parse_var_dims(var_dims, var_names)
        var_coords = _parse_var_coords(var_coords)

    # Generate results
    results = _case_runner(fn, fn_args, cases,
                           constants={**constants, **resources},
                           **case_runner_settings)

    # Convert to xarray.Dataset
    ds = _cases_to_ds(results, fn_args, cases,
                      var_names=var_names,
                      var_dims=var_dims,
                      var_coords=var_coords,
                      constants=constants,
                      attrs=attrs,
                      add_to_ds=add_to_ds,
                      overwrite=overwrite)
    return ds


# --------------------------------------------------------------------------- #
# Update or add new values                                                    #
# --------------------------------------------------------------------------- #

def find_missing_cases(ds, ignore_dims=None, show_progbar=False):
    """Find all cases in a dataset with missing data.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset in which to find missing data
    ignore_dims : set (optional)
        internal variable dimensions (i.e. to ignore)
    show_progbar : bool (optional)
        Show the current progress

    Returns
    -------
    missing_fn_args, missing_cases :
        Function arguments and missing cases.
    """
    # Parse ignore_dims
    ignore_dims = ({ignore_dims} if isinstance(ignore_dims, str) else
                   set(ignore_dims) if ignore_dims else set())

    # Find all configurations
    fn_args = tuple(coo for coo in ds.dims if coo not in ignore_dims)
    var_names = tuple(ds.data_vars)
    all_cases = itertools.product(*(ds[arg].data for arg in fn_args))

    # Only return those corresponding to all missing data
    def gen_missing_list():
        for case in progbar(all_cases, disable=not show_progbar):
            sub_ds = ds.loc[dict(zip(fn_args, case))]
            if all(sub_ds[v].isnull().all() for v in var_names):
                yield case

    return fn_args, tuple(gen_missing_list())


def fill_missing_cases(ds, fn, var_names,
                       var_dims=None,
                       var_coords=None,
                       constants=None,
                       resources=None,
                       **case_runner_settings):
    """ Take a dataset and function etc. and fill its missing data in

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to analyse and fill
    fn : callable
        Function to use to fill missing cases
    var_names : tuple
        Output variable names of function
    var_dims : dict
        Output variabe named dimensions of function
    var_coords : dict
        Dictionary of coords for output dims
    case_runner_settings:
        Supplied to :func:`~xyzpy.case_runner`.

    Returns
    -------
    xarray.Dataset
    """
    var_names = _parse_var_names(var_names)
    var_dims = _parse_var_dims(var_dims, var_names)
    var_coords = _parse_var_coords(var_coords)
    constants = _parse_constants(constants)
    resources = _parse_resources(resources)

    # Gather all internal dimensions
    ignore_dims = set(concat(var_dims.values()))

    # Find missing cases
    fn_args, missing_cases = find_missing_cases(ds, ignore_dims=ignore_dims)

    # Generate missing results
    results = _case_runner(fn, fn_args, missing_cases,
                           constants={**constants, **resources},
                           **case_runner_settings)

    # Add to dataset
    return _cases_to_ds(results, fn_args, missing_cases,
                        var_names=var_names,
                        var_dims=var_dims,
                        var_coords=var_coords,
                        add_to_ds=ds)



"""
"""
import functools
import xarray as xr

from .utils import (
    unzip,
    progbar,
    Timer,
)
from .gen.combo_runner import (
    combo_runner,
    combo_runner_to_ds
)
from .gen.case_runner import (
    case_runner,
    find_union_coords,
    all_missing_ds,
    case_runner_to_ds,
    find_missing_cases,
    fill_missing_cases
)
from .gen.batch import (
    Crop,
    grow,
)
from .gen.farming import (
    Runner,
    Harvester
)
from .manage import (
    cache_to_disk,
    save_ds,
    load_ds,
    trimna,
    sort_dims,
    check_runs,
    auto_xyz_ds,
    merge_sync_conflict_datasets,
    post_fix,
)
from .signal import (
    xr_diff_fornberg,
    xr_diff_u,
    xr_diff_u_err,
    xr_interp,
    xr_interp_pchip,
    xr_filter_wiener,
    xr_filtfilt_butter,
    xr_filtfilt_bessel,
    xr_unispline,
    xr_polyfit,
)
from .plot.color import (
    convert_colors,
)
# Making static plots with matplotlib
from .plot.plotter_matplotlib import (
    LinePlot,
    lineplot,
    AutoLinePlot,
    auto_lineplot,
    Scatter,
    scatter,
    AutoScatter,
    auto_scatter,
    Histogram,
    histogram,
    AutoHistogram,
    auto_histogram,
    HeatMap,
    heatmap,
    AutoHeatMap,
    auto_heatmap,
    visualize_matrix
)
# Making interactive plots with bokeh
from .plot.plotter_bokeh import (
    ilineplot,
    auto_ilineplot,
    iscatter,
    auto_iscatter,
    iheatmap,
    auto_iheatmap,
)


# versioneer
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


__all__ = [
    "Runner",
    "Harvester",
    "combo_runner",
    "combo_runner_to_ds",
    "case_runner",
    "find_union_coords",
    "all_missing_ds",
    "case_runner_to_ds",
    "find_missing_cases",
    "fill_missing_cases",
    "Crop",
    "grow",
    "cache_to_disk",
    "save_ds",
    "load_ds",
    "trimna",
    "sort_dims",
    "check_runs",
    "merge_sync_conflict_datasets",
    "auto_xyz_ds",
    "convert_colors",
    "LinePlot",
    "lineplot",
    "auto_lineplot",
    "AutoLinePlot",
    "Scatter",
    "scatter",
    "AutoScatter",
    "auto_scatter",
    "Histogram",
    "histogram",
    "AutoHistogram",
    "auto_histogram",
    "HeatMap",
    "heatmap",
    "AutoHeatMap",
    "auto_heatmap",
    "ilineplot",
    "auto_ilineplot",
    "iscatter",
    "auto_iscatter",
    "iheatmap",
    "auto_iheatmap",
    "visualize_matrix",
    "unzip",
    "progbar",
    "Timer",
    "xr_diff_fornberg",
    "xr_diff_u",
    "xr_diff_u_err",
    "xr_interp",
    "xr_interp_pchip",
    "xr_filter_wiener",
    "xr_filtfilt_butter",
    "xr_filtfilt_bessel",
    "xr_unispline",
    "xr_polyfit",
]


class XYZPY(object):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    # ------------------------------- Plotting ------------------------------ #

    @functools.wraps(LinePlot)
    def LinePlot(self, *args, **kwargs):
        return LinePlot(self._obj, *args, **kwargs)

    @functools.wraps(lineplot)
    def lineplot(self, *args, **kwargs):
        return lineplot(self._obj, *args, **kwargs)

    @functools.wraps(Scatter)
    def Scatter(self, *args, **kwargs):
        return Scatter(self._obj, *args, **kwargs)

    @functools.wraps(scatter)
    def scatter(self, *args, **kwargs):
        return scatter(self._obj, *args, **kwargs)

    @functools.wraps(Histogram)
    def Histogram(self, *args, **kwargs):
        return Histogram(self._obj, *args, **kwargs)

    @functools.wraps(histogram)
    def histogram(self, *args, **kwargs):
        return histogram(self._obj, *args, **kwargs)

    @functools.wraps(HeatMap)
    def HeatMap(self, *args, **kwargs):
        return HeatMap(self._obj, *args, **kwargs)

    @functools.wraps(heatmap)
    def heatmap(self, *args, **kwargs):
        return heatmap(self._obj, *args, **kwargs)

    @functools.wraps(ilineplot)
    def ilineplot(self, *args, **kwargs):
        return ilineplot(self._obj, *args, **kwargs)

    @functools.wraps(iscatter)
    def iscatter(self, *args, **kwargs):
        return iscatter(self._obj, *args, **kwargs)

    @functools.wraps(iheatmap)
    def iheatmap(self, *args, **kwargs):
        return iheatmap(self._obj, *args, **kwargs)

    # ----------------------------- Processing ------------------------------ #

    @functools.wraps(trimna)
    def trimna(self):
        return trimna(self._obj)

    @functools.wraps(post_fix)
    def post_fix(self, postfix):
        return post_fix(self._obj, postfix)

    @functools.wraps(xr_diff_fornberg)
    def diff_fornberg(self, dim, ix=100, order=1, mode='points', window=5):
        return xr_diff_fornberg(self._obj, dim=dim, ix=ix, order=order,
                                mode=mode, window=window)

    @functools.wraps(xr_diff_u)
    def diff_u(self, dim):
        return xr_diff_u(self._obj, dim=dim)

    @functools.wraps(xr_diff_u_err)
    def diff_u_err(self, dim):
        return xr_diff_u_err(self._obj, dim=dim)

    @functools.wraps(xr_interp)
    def interp(self, dim, ix=100, order=3):
        return xr_interp(self._obj, dim=dim, ix=ix, order=order)

    @functools.wraps(xr_interp_pchip)
    def interp_pchip(self, dim, ix=100):
        return xr_interp_pchip(self._obj, dim=dim, ix=ix)

    @functools.wraps(xr_filter_wiener)
    def filter_wiener(self, dim, mysize=5, noise=1e-2):
        return xr_filter_wiener(self._obj, dim=dim, mysize=mysize, noise=noise)

    @functools.wraps(xr_filtfilt_butter)
    def filtfilt_butter(self, dim, N=2, Wn=0.4):
        return xr_filtfilt_butter(self._obj, dim=dim, N=N, Wn=Wn)

    @functools.wraps(xr_filtfilt_bessel)
    def filtfilt_bessel(self, dim, N=2, Wn=0.4):
        return xr_filtfilt_bessel(self._obj, dim=dim, N=N, Wn=Wn)

    @functools.wraps(xr_unispline)
    def unispline(self, dim, err=None, num_knots=11, ix=None):
        return xr_unispline(self._obj, dim=dim, err=err,
                            num_knots=num_knots, ix=ix)

    @functools.wraps(xr_polyfit)
    def polyfit(self, dim, ix=None, deg=0.5, poly='chebyshev'):
        return xr_polyfit(self._obj, dim=dim, ix=ix, deg=deg, poly=poly)


xr.register_dataarray_accessor('xyz')(XYZPY)
xr.register_dataset_accessor('xyz')(XYZPY)
